import copy
import sys
from datetime import timedelta, date

import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from sodapy import Socrata
import pandas as pd
import numpy as np

from metrics.metrics import Metrics
from models.models import Models


def get_data():
    # Unauthenticated client only works with public data sets. Note 'None'
    # in place of application token, and no username or password:
    client = Socrata("analisi.transparenciacatalunya.cat",
                     "KttyqTxvzEszYeIrufVocTlgr",
                     username="albert.sola9@gmail.com",
                     password="m3wK5EihhAmsUMi")

    # First 300000 results, returned as JSON from API / converted to Python list of
    # dictionaries by sodapy.
    results = client.get("c7sd-zy9j", limit=300000)  # update limit as days go by

    # Convert to pandas DataFrame
    results_df = pd.DataFrame.from_records(results)
    results_df['data'] = pd.to_datetime(results_df['data'])
    results_df = results_df[['data', 'casos_confirmat']].astype({'casos_confirmat': 'int64'})
    return results_df


def parse_data(results_df):
    #Group by days by adding
    casos_diaris = results_df.groupby(by='data').sum()

    #Add the cumulative cases to the dataframe
    casos_diaris['casos_cumulatius'] = casos_diaris['casos_confirmat'].cumsum()
    casos_diaris.join(results_df['data'])
    return casos_diaris


def plot_dates():
    #Method to plot the important dates on all of our graphs
    l1 = plt.axvline(date.fromisoformat("2020-05-02"), color='g')
    l2 = plt.axvline(date.fromisoformat("2020-10-16"), color='r')
    l3 = plt.axvline(date.fromisoformat("2020-12-14"), color='g')
    l4 = plt.axvline(date.fromisoformat("2021-01-07"), color='r')
    l5 = plt.axvline(date.fromisoformat("2021-04-26"), color='g')
    plt.legend([(l1, l3, l5), (l2, l4)], ['reopening', 'restrictions'])


def predict_SEIRS(start_date, days, casos_diaris):
    prediction_SEIRS = []

    for i in range(start_date, len(casos_diaris) - days):
        print("SEIRS ", i)
        xdata = np.arange(i - 30, i)
        ydata = copy.copy(casos_diaris['casos_confirmat'].rolling(5).mean().values[i - 30:i])

        initial_pred_SEIRS = [25000, 0.2, 0.3, 0.1, 0.01]
        min_fit = curve_fit(models.SEIRS, xdata, ydata, p0=initial_pred_SEIRS, max_nfev=100000,
                            bounds=([0, 0, 0, 0, 0], [1000000, 1, 1, 1, 1]), ftol=1e-7)

        prediction = copy.copy(casos_diaris['casos_cumulatius'].values[i - 1:i]) + np.cumsum(
            models.SEIRS(i - 30, min_fit[0][0], min_fit[0][1], min_fit[0][2], min_fit[0][3], min_fit[0][4],
                         days=30 + days)[30:30 + days])
        prediction_SEIRS.append(prediction[30 + days])

    date_list = pd.date_range(start=date.fromisoformat("2020-02-29") + timedelta(days + start_date),
                              end=date.fromisoformat("2020-02-29") + timedelta(len(casos_diaris) - 1))
    d = {'data': date_list, 'casos cumulatius': casos_diaris['casos_cumulatius'].values[start_date + days:],
         'prediction SEIRS': prediction_SEIRS}
    prediction_df = pd.DataFrame(data=d).groupby(by='data').sum()

    plt.figure()
    prediction_df.plot(legend=True)
    plt.title('Predict SEIRS for ' + str(days) + ' days')
    plt.ylabel("Cumulative cases")
    plt.xlabel("Date")
    plot_dates()
    plt.autoscale()
    plt.savefig('images/SEIRS14.png', bbox_inches='tight')
    plt.show()

    return prediction_SEIRS


def predict_SEIR(start_date, days, casos_diaris):
    prediction_SEIR = []

    for i in range(start_date, len(casos_diaris) - days):
        print("SEIR ", i)
        xdata = np.arange(i - 30, i)
        ydata = copy.copy(casos_diaris['casos_confirmat'].rolling(5).mean().values[i - 30:i])

        initial_pred_SEIR = [25000, 0.2, 0.3, 0.1]
        min_fit = curve_fit(models.SEIR, xdata, ydata, p0=initial_pred_SEIR, max_nfev=100000,
                            bounds=([0, 0, 0, 0], [1000000, 1, 1, 1]), ftol=1e-7)

        prediction = copy.copy(casos_diaris['casos_cumulatius'].values[i - 1:i]) + np.cumsum(
            models.SEIR(i - 30, min_fit[0][0], min_fit[0][1], min_fit[0][2], min_fit[0][3], days=30 + days)[
            30:30 + days])
        prediction_SEIR.append(prediction[30 + days])

    date_list = pd.date_range(start=date.fromisoformat("2020-02-29") + timedelta(days + start_date),
                              end=date.fromisoformat("2020-02-29") + timedelta(len(casos_diaris) - 1))
    d = {'data': date_list, 'casos cumulatius': casos_diaris['casos_cumulatius'].values[start_date + days:],
         'prediction SEIR': prediction_SEIR}
    prediction_df = pd.DataFrame(data=d).groupby(by='data').sum()

    plt.figure()
    prediction_df.plot(legend=True)
    plt.title('Predict SEIR for ' + str(days) + ' days')
    plt.ylabel("Cumulative cases")
    plt.xlabel("Date")
    plot_dates()
    plt.autoscale()
    plt.savefig('images/SEIR14.png', bbox_inches='tight')
    plt.show()

    return prediction_SEIR


def predict_SIR(start_date, days, casos_diaris):
    prediction_SIR = []

    for i in range(start_date, len(casos_diaris) - days):
        print("SIR ", i)
        xdata = np.arange(i - 30, i)
        ydata = copy.copy(casos_diaris['casos_confirmat'].rolling(5).mean().values[i - 30:i])

        initial_pred_SIR = [0.2, 0.1]

        min_fit = curve_fit(models.SIR, xdata, ydata, p0=initial_pred_SIR, max_nfev=100000, bounds=([0, 0], [1, 1]),
                            ftol=1e-7)

        prediction = copy.copy(casos_diaris['casos_cumulatius'].values[i - 1:i]) + np.cumsum(
            models.SIR(i - 30, min_fit[0][0], min_fit[0][1], days=30 + days)[30:30 + days])
        prediction_SIR.append(prediction[30 + days])

    date_list = pd.date_range(start=date.fromisoformat("2020-02-29") + timedelta(days + start_date),
                              end=date.fromisoformat("2020-02-29") + timedelta(len(casos_diaris) - 1))
    d = {'data': date_list, 'casos cumulatius': casos_diaris['casos_cumulatius'].values[start_date + days:],
         'prediction SIR': prediction_SIR}
    prediction_df = pd.DataFrame(data=d).groupby(by='data').sum()

    plt.figure()
    prediction_df.plot(legend=True)
    plt.title('Predict SIR for ' + str(days) + ' days')
    plt.ylabel("Cumulative cases")
    plt.xlabel("Date")
    plot_dates()
    plt.autoscale()
    plt.savefig('images/SIR14.png', bbox_inches='tight')
    plt.show()

    return prediction_SIR


def predict_gompertz(start_date, days, casos_diaris):
    prediction_gompertz = []
    a = []
    k = []
    mu = []

    for i in range(start_date, len(casos_diaris) - days):
        print("Gompertz", i)
        ydata = copy.copy(casos_diaris['casos_cumulatius'].values[i - 30:i])
        k0 = copy.copy(casos_diaris['casos_cumulatius'].values[i - 1:i])
        ini_guess = [0.01, k0]

        xdata = np.arange(i - 30, i)

        min_fit = curve_fit(models.gompertz_model, xdata, ydata, p0=ini_guess, max_nfev=100000,
                            bounds=([0, 0], [1, 1000000]), ftol=1e-7)
        a.append(min_fit[0][0])
        k.append(min_fit[0][1])
        mu.append(min_fit[0][0]*np.log(min_fit[0][1]/copy.copy(casos_diaris['casos_cumulatius'].values[start_date - 30:start_date - 29])))

        prediction_gompertz.append(int(models.gompertz_model(i + days, min_fit[0][0], min_fit[0][1])))

    date_list = pd.date_range(start=date.fromisoformat("2020-02-29") + timedelta(days + start_date),
                              end=date.fromisoformat("2020-02-29") + timedelta(len(casos_diaris) - 1))
    d = {'data': date_list, 'casos cumulatius': casos_diaris['casos_cumulatius'].values[start_date + days:],
         'prediction gompertz': prediction_gompertz}
    # print(len(date_list), len(casos_diaris['casos_cumulatius'].values[start_date + days:]), len(prediction_gompertz))
    prediction_df = pd.DataFrame(data=d).groupby(by='data').sum()
    # print(prediction_df)

    plt.figure()
    prediction_df.plot(legend=True)
    plt.title('Predict Gompertz for ' + str(days) + ' days')
    plt.ylabel("Cumulative cases")
    plt.xlabel("Date")
    plot_dates()
    plt.autoscale()
    plt.savefig('images/gompertz14.png', bbox_inches='tight')
    plt.show()

    plt.figure()
    plt.plot(date_list, a)
    plt.title('Variation of parameter a')
    plt.ylabel("a")
    plt.xlabel("Date")
    plot_dates()
    plt.autoscale()
    plt.savefig('images/gompertz14a.png', bbox_inches='tight')
    plt.show()

    plt.figure()
    plt.plot(date_list, k)
    plt.title('Variation of parameter K')
    plt.ylabel("K")
    plt.xlabel("Date")
    plot_dates()
    plt.autoscale()
    plt.savefig('images/gompertz14k.png', bbox_inches='tight')
    plt.show()

    plt.figure()
    plt.plot(date_list, mu)
    plt.title('Variation of parameter μ')
    plt.ylabel("μ")
    plt.xlabel("Date")
    plot_dates()
    plt.autoscale()
    plt.savefig('images/gompertz14mu.png', bbox_inches='tight')
    plt.show()

    return prediction_gompertz


def predict_average(prediction_df):
    prediction_df['prediction average'] = prediction_df[['prediction SEIR', 'prediction gompertz']].mean(axis=1)
    plt.figure()
    prediction_df.plot(x="data", y=["casos cumulatius", "prediction average"])
    plt.title('Predict Average of Gompertz and SEIR for ' + str(14) + ' days')
    plt.xlabel("Date")
    plt.ylabel("Cumulative cases")
    plot_dates()
    plt.savefig('images/average14.png', bbox_inches='tight')
    plt.show()


def plot_metrics(prediction_df):

    metrics = Metrics(prediction_df)
    r_squared_gompertz, r_squared_SIR, r_squared_SEIR, r_squared_SEIRS, r_squared_average = metrics.r_squared()
    error_gompertz, error_SIR, error_SEIR, error_SEIRS, error_average = metrics.error_rate()
    mean_gompertz, mean_SIR, mean_SEIR, mean_SEIRS, mean_average, real_mean = metrics.mean_diff()

    plt.figure()
    errors = ["Gompertz", "SIR", "SEIR", "SEIRS", "Average"]
    values = [error_gompertz, error_SIR, error_SEIR, error_SEIRS, error_average]
    plt.bar(errors, values)
    plt.title("Cumulative relative error rates of the different methods")
    plt.xlabel("Model")
    plt.ylabel("Error rate %")
    plt.savefig("images/Error rates.png", bbox_inches='tight')

    print("R^2: ", r_squared_gompertz, r_squared_SIR, r_squared_SEIR, r_squared_SEIRS, r_squared_average)
    print("Error Rates", error_gompertz, error_SIR, error_SEIR, error_SEIRS, error_average)
    print("Mean differences:", mean_gompertz, mean_SIR, mean_SEIR, mean_SEIRS, mean_average, "Real Peak difference: ",
          real_mean)


if __name__ == "__main__":
    results_df = get_data()
    casos_diaris = parse_data(results_df)

    start_date = 45
    models = Models(casos_diaris, start_date)

    prediction_SEIRS = predict_SEIRS(start_date, 14, casos_diaris)
    prediction_SEIR = predict_SEIR(start_date, 14, casos_diaris)
    prediction_SIR = predict_SIR(start_date, 14, casos_diaris)
    prediction_gompertz = predict_gompertz(start_date, 14, casos_diaris)

    date_list = pd.date_range(start=date.fromisoformat("2020-02-29") + timedelta(14 + start_date),
                              end=date.fromisoformat("2020-02-29") + timedelta(len(casos_diaris) - 1))
    d = {'data': date_list, 'casos cumulatius': casos_diaris['casos_cumulatius'].values[start_date + 14:],
         'prediction gompertz': prediction_gompertz, 'prediction SIR': prediction_SIR,
         'prediction SEIR': prediction_SEIR,
         'prediction SEIRS': prediction_SEIRS}

    prediction_df = pd.DataFrame(data=d)

    predict_average(prediction_df)

    plot_metrics(prediction_df)
    sys.exit(0)
