from datetime import date

import numpy as np
import scipy.stats
import pandas as pd
import matplotlib.pyplot as plt


class Metrics(object):
    prediction_df = 0

    def __init__(self, prediction_df):
        self.prediction_df = prediction_df
        self.prediction_df['data'] = pd.to_datetime(self.prediction_df['data'])

    def error_rate(self):
        self.prediction_df['error rate gompertz'] = np.abs(
            self.prediction_df['casos cumulatius'] - self.prediction_df['prediction gompertz'])
        self.prediction_df['error rate SIR'] = np.abs(
            self.prediction_df['casos cumulatius'] - self.prediction_df['prediction SIR'])
        self.prediction_df['error rate SEIR'] = np.abs(
            self.prediction_df['casos cumulatius'] - self.prediction_df['prediction SEIR'])
        self.prediction_df['error rate SEIRS'] = np.abs(
            self.prediction_df['casos cumulatius'] - self.prediction_df['prediction SEIRS'])
        self.prediction_df['error rate average'] = np.abs(
            self.prediction_df['casos cumulatius'] - self.prediction_df['prediction average'])

        plt.figure()
        self.prediction_df.plot.area(x='data', y='error rate gompertz')
        mean = plt.axhline(self.prediction_df['error rate gompertz'].mean(),color='y')
        plt.ylabel("Difference")
        plt.xlabel("Date")
        plt.title("Daily error of Gompertz model")
        l1 = plt.axvline(date.fromisoformat("2020-05-02"), color='g')
        l2 = plt.axvline(date.fromisoformat("2020-10-16"), color='r')
        l3 = plt.axvline(date.fromisoformat("2020-12-14"), color='g')
        l4 = plt.axvline(date.fromisoformat("2021-01-07"), color='r')
        l5 = plt.axvline(date.fromisoformat("2021-04-26"), color='g')
        plt.legend([(l1, l3, l5), (l2, l4),(mean)], ['reopening', 'restrictions','mean error'])
        plt.autoscale()
        plt.savefig('images/Error rate gompertz',bbox_inches='tight')

        plt.figure()
        self.prediction_df.plot.area(x='data', y='error rate SIR')
        mean = plt.axhline(self.prediction_df['error rate SIR'].mean(),color='y')
        plt.title("Daily error of SIR model")
        plt.ylabel("Difference")
        plt.xlabel("Date")
        l1 = plt.axvline(date.fromisoformat("2020-05-02"), color='g')
        l2 = plt.axvline(date.fromisoformat("2020-10-16"), color='r')
        l3 = plt.axvline(date.fromisoformat("2020-12-14"), color='g')
        l4 = plt.axvline(date.fromisoformat("2021-01-07"), color='r')
        l5 = plt.axvline(date.fromisoformat("2021-04-26"), color='g')
        plt.legend([(l1, l3, l5), (l2, l4), mean], ['reopening', 'restrictions', 'mean error'])
        plt.autoscale()
        plt.savefig('images/Error rate SIR',bbox_inches='tight')

        plt.figure()
        self.prediction_df.plot.area(x='data', y='error rate SEIR')
        mean = plt.axhline(self.prediction_df['error rate SEIR'].mean(),color='y')
        plt.title("Daily error of SEIR model")
        plt.xlabel("Date")
        plt.ylabel("Difference")
        l1 = plt.axvline(date.fromisoformat("2020-05-02"), color='g')
        l2 = plt.axvline(date.fromisoformat("2020-10-16"), color='r')
        l3 = plt.axvline(date.fromisoformat("2020-12-14"), color='g')
        l4 = plt.axvline(date.fromisoformat("2021-01-07"), color='r')
        l5 = plt.axvline(date.fromisoformat("2021-04-26"), color='g')
        plt.legend([(l1, l3, l5), (l2, l4), mean], ['reopening', 'restrictions', 'mean error'])
        plt.autoscale()
        plt.savefig('images/Error rate SEIR',bbox_inches='tight')

        plt.figure()
        self.prediction_df.plot.area(x='data', y='error rate SEIRS')
        mean = plt.axhline(self.prediction_df['error rate SEIRS'].mean(),color='y')
        plt.title("Daily error of SEIRS model")
        plt.xlabel("Date")
        plt.ylabel("Difference")
        l1 = plt.axvline(date.fromisoformat("2020-05-02"), color='g')
        l2 = plt.axvline(date.fromisoformat("2020-10-16"), color='r')
        l3 = plt.axvline(date.fromisoformat("2020-12-14"), color='g')
        l4 = plt.axvline(date.fromisoformat("2021-01-07"), color='r')
        l5 = plt.axvline(date.fromisoformat("2021-04-26"), color='g')
        plt.legend([(l1, l3, l5), (l2, l4), mean], ['reopening', 'restrictions', 'mean error'])
        plt.autoscale()
        plt.savefig('images/Error rate SEIRS',bbox_inches='tight')

        plt.figure()
        self.prediction_df.plot.area(x='data', y='error rate average')
        mean = plt.axhline(self.prediction_df['error rate average'].mean(),color='y')
        plt.title("Daily error of average of SEIR and Gompertz model")
        plt.xlabel("Date")
        plt.ylabel("Difference")
        l1 = plt.axvline(date.fromisoformat("2020-05-02"), color='g')
        l2 = plt.axvline(date.fromisoformat("2020-10-16"), color='r')
        l3 = plt.axvline(date.fromisoformat("2020-12-14"), color='g')
        l4 = plt.axvline(date.fromisoformat("2021-01-07"), color='r')
        l5 = plt.axvline(date.fromisoformat("2021-04-26"), color='g')
        plt.legend([(l1, l3, l5), (l2, l4), mean], ['reopening', 'restrictions', 'mean error'])
        plt.autoscale()
        plt.savefig('images/Error rate average',bbox_inches='tight')

        return self.prediction_df['error rate gompertz'].sum() / self.prediction_df['casos cumulatius'].sum()*100, \
               self.prediction_df['error rate SIR'].sum() / self.prediction_df['casos cumulatius'].sum()*100, \
               self.prediction_df['error rate SEIR'].sum() / self.prediction_df['casos cumulatius'].sum()*100, \
               self.prediction_df['error rate SEIRS'].sum() / self.prediction_df['casos cumulatius'].sum()*100, \
               self.prediction_df['error rate average'].sum() / self.prediction_df['casos cumulatius'].sum()*100

    def mean_diff(self):
        prediction_summed = self.prediction_df.groupby(by='data').sum()
        prediction_summed['peak rate gompertz'] = (
            np.abs(prediction_summed['prediction gompertz'] - prediction_summed['prediction gompertz'].shift(1)))
        prediction_summed['peak rate SIR'] = np.abs(
            prediction_summed['prediction SIR'] - prediction_summed['prediction SIR'].shift(1))
        prediction_summed['peak rate SEIR'] = np.abs(
            prediction_summed['prediction SEIR'] - prediction_summed['prediction SEIR'].shift(1))
        prediction_summed['peak rate SEIRS'] = np.abs(
            prediction_summed['prediction SEIRS'] - prediction_summed['prediction SEIRS'].shift(1))
        prediction_summed['peak rate average'] = np.abs(
            prediction_summed['prediction average'] - prediction_summed['prediction average'].shift(1))
        prediction_summed['real peak rate'] = np.abs(
            prediction_summed['casos cumulatius'] - prediction_summed['casos cumulatius'].shift(1))
        return prediction_summed['peak rate gompertz'].mean(), \
               prediction_summed['peak rate SIR'].mean(), \
               prediction_summed['peak rate SEIR'].mean(), \
               prediction_summed['peak rate SEIRS'].mean(), \
               prediction_summed['peak rate average'].mean(), \
               prediction_summed['real peak rate'].mean()

    def r_squared(self):
        return scipy.stats.linregress(
            self.prediction_df[['casos cumulatius', 'prediction gompertz']].to_numpy()).rvalue ** 2, \
               scipy.stats.linregress(
                   self.prediction_df[['casos cumulatius', 'prediction SIR']].to_numpy()).rvalue ** 2, \
               scipy.stats.linregress(
                   self.prediction_df[['casos cumulatius', 'prediction SEIR']].to_numpy()).rvalue ** 2, \
               scipy.stats.linregress(
                   self.prediction_df[['casos cumulatius', 'prediction SEIRS']].to_numpy()).rvalue ** 2, \
               scipy.stats.linregress(
                   self.prediction_df[['casos cumulatius', 'prediction average']].to_numpy()).rvalue ** 2
