import pandas as pd
import numpy as np
import copy

from models.EpiModel import EpiModel


class Models(object):
    casos_diaris = 0
    start_date=0

    def __init__(self, casos_diaris,start_date):
        self.casos_diaris = casos_diaris
        self.start_date = start_date

    def gompertz_model(self, x, a,k):
        n0 = copy.copy(self.casos_diaris['casos_cumulatius'].values[self.start_date - 30:self.start_date - 29])
        return k * np.exp(-np.log(k / n0) * np.exp(-x * a))

    def SEIRS(self, t, E, beta, mu, epsilon, rho, days=30):
        SEIRS = EpiModel()
        SEIRS.add_interaction('S', 'E', 'I', beta)
        SEIRS.add_spontaneous('E', 'I', epsilon)
        SEIRS.add_spontaneous('I', 'R', mu)
        SEIRS.add_spontaneous('R', 'S', rho)

        if not isinstance(t, int):
            t = int(t[0])

        I0 = copy.copy(self.casos_diaris['casos_confirmat'].rolling(5).mean().values[t:t + 1])
        R0 = copy.copy(self.casos_diaris['casos_cumulatius'].values[t - 21:t - 20]) if t - 21 > 0 else 0
        infectats_totals = copy.copy(self.casos_diaris['casos_cumulatius'].values[t:t + 1])
        SEIRS.integrate(days, S=7556000 - infectats_totals-E-R0, I=I0, R=R0, E=E)

        return SEIRS.I

    def SEIR(self, t, E, beta, mu, epsilon, days=30):

        SEIR = EpiModel()
        SEIR.add_interaction('S', 'E', 'I', beta)
        SEIR.add_spontaneous('E', 'I', epsilon)
        SEIR.add_spontaneous('I', 'R', mu)

        if not isinstance(t, int):
            t = int(t[0])

        I0 = copy.copy(self.casos_diaris['casos_confirmat'].rolling(7).mean().values[t:t + 1])
        R0 = copy.copy(self.casos_diaris['casos_cumulatius'].values[t - 21:t - 20]) if t - 21 > 0 else 0
        infectats_totals = copy.copy(self.casos_diaris['casos_cumulatius'].values[t:t + 1])
        SEIR.integrate(days, S=7556000 - infectats_totals-E-R0, I=I0, R=R0, E=E)

        return SEIR.I

    def SIR(self, t, beta, mu, days=30):
        SIR = EpiModel()
        SIR.add_interaction('S', 'I', 'I', beta)
        SIR.add_spontaneous('I', 'R', mu)

        if not isinstance(t, int):
            t = int(t[0])

        I0 = copy.copy(self.casos_diaris['casos_confirmat'].rolling(5).mean().values[t:t + 1])
        R0 = copy.copy(self.casos_diaris['casos_cumulatius'].values[t - 21:t - 20]) if t - 21 > 0 else 0
        infectats_totals = copy.copy(self.casos_diaris['casos_cumulatius'].values[t:t + 1])
        SIR.integrate(days, S=7556000 - infectats_totals, I=I0, R=R0)

        return SIR.I
